/*
 * Copyright (C) 2008 Emweb bvba, Kessel-Lo, Belgium.
 *
 * See the LICENSE file for terms of use.
 */
#include <boost/lexical_cast.hpp>

#include "Wt/WApplication"
#include "Wt/WContainerWidget"
#include "Wt/WLogger"
#include "Wt/WWebWidget"
#include "Wt/WJavaScript"

#include "DomElement.h"
#include "WebRenderer.h"
#include "WebSession.h"
#include "WebSession.h"
#include "WtException.h"
#include "Utils.h"
#include "XSSFilter.h"

using namespace Wt;

std::vector<WWidget *> WWebWidget::emptyWidgetList_;

WWebWidget::TransientImpl::TransientImpl()
{ }

WWebWidget::TransientImpl::~TransientImpl()
{
  for (unsigned i = 0; i < childRemoveChanges_.size(); ++i)
    delete childRemoveChanges_[i];
}

WWebWidget::LayoutImpl::LayoutImpl()
  : positionScheme_(Static),
#ifndef WT_TARGET_JAVA
    floatSide_(static_cast<Side>(0)),
    clearSides_(0),
#endif // WT_TARGET_JAVA
    popup_(false),
    verticalAlignment_(AlignBaseline),
    marginsChanged_(false)
{ 
  for (unsigned i = 0; i < 4; ++i) {
#ifdef WT_TARGET_JAVA
    offsets_[i] = WLength::Auto;
#endif // WT_TARGET_JAVA
    margin_[i] = WLength(0);
  }
}

WWebWidget::LookImpl::LookImpl()
  : decorationStyle_(0),
    toolTip_(0),
    styleClassChanged_(false),
    toolTipChanged_(false)
{ }

WWebWidget::LookImpl::~LookImpl()
{
  delete decorationStyle_;
  delete toolTip_;
}

WWebWidget::OtherImpl::OtherImpl()
  : attributes_(0),
    attributesSet_(0),
    id_(0),
    dropSignal_(0),
    acceptedDropMimeTypes_(0)
{ }

WWebWidget::OtherImpl::~OtherImpl()
{
  delete attributes_;
  delete attributesSet_;
  delete dropSignal_;
  delete acceptedDropMimeTypes_;
  delete id_;
}

WWebWidget::WWebWidget(WContainerWidget *parent)
  : WWidget(parent),
    width_(0),
    height_(0),
    transientImpl_(0),
    layoutImpl_(0),
    lookImpl_(0),
    otherImpl_(0),
    children_(0)
{
  flags_.set(BIT_INLINE);

  if (parent)
    parent->addWidget(this);
}

void WWebWidget::setFormObject(bool how)
{
  flags_.set(BIT_FORM_OBJECT, how);

  WApplication::instance()
    ->session()->renderer().updateFormObjects(this, false);
}

void WWebWidget::setId(const std::string& id)
{
  if (!otherImpl_)
    otherImpl_ = new OtherImpl();

  if (!otherImpl_->id_)
    otherImpl_->id_ = new std::string();

  *otherImpl_->id_ = id;
}

const std::string WWebWidget::formName() const
{
  if (otherImpl_ && otherImpl_->id_)
    return *otherImpl_->id_;
  else
    return WWidget::formName();
}

void WWebWidget::repaint(WFlags<RepaintFlag> flags)
{
  WWidget::askRerender();

#ifndef WT_TARGET_JAVA
  flags_ |= (int)flags;
#else
  if (flags & RepaintPropertyIEMobile)
    flags_.set(BIT_REPAINT_PROPERTY_IEMOBILE);
  if (flags & RepaintPropertyAttribute)
    flags_.set(BIT_REPAINT_PROPERTY_ATTRIBUTE);
  if (flags & RepaintInnerHtml)
    flags_.set(BIT_REPAINT_INNER_HTML);
#endif // WT_TARGET_JAVA
}

void WWebWidget::renderOk()
{
  WWidget::renderOk();

#ifndef WT_TARGET_JAVA
  flags_ &= ~(int)RepaintAll;
#else // WT_TARGET_JAVA
  flags_.reset(BIT_REPAINT_PROPERTY_IEMOBILE);
  flags_.reset(BIT_REPAINT_PROPERTY_ATTRIBUTE);
  flags_.reset(BIT_REPAINT_INNER_HTML);
#endif // WT_TARGET_JAVA
}

void WWebWidget::signalConnectionsChanged()
{
  repaint(RepaintPropertyAttribute);
}

WWebWidget::~WWebWidget()
{
  flags_.set(BIT_BEING_DELETED);

  if (flags_.test(BIT_FORM_OBJECT))
    WApplication::instance()
      ->session()->renderer().updateFormObjects(this, false);

  setParent(0);

  delete width_;
  delete height_;

  if (children_) {
    while (children_->size())
      delete (*children_)[0];
    delete children_;
  }

  delete transientImpl_;
  delete layoutImpl_;
  delete lookImpl_;
  delete otherImpl_;
}

WCssDecorationStyle& WWebWidget::decorationStyle()
{
  if (!lookImpl_)
    lookImpl_ = new LookImpl();

  if (!lookImpl_->decorationStyle_) {
    lookImpl_->decorationStyle_ = new WCssDecorationStyle();
    lookImpl_->decorationStyle_->setWebWidget(this);
  }

  return *lookImpl_->decorationStyle_;
}

void WWebWidget::setDecorationStyle(const WCssDecorationStyle& style)
{
#ifndef WT_TARGET_JAVA
  decorationStyle() = style;
#else
  if (!lookImpl_)
    lookImpl_ = new LookImpl();

  lookImpl_->decorationStyle_ = &style;
#endif // WT_TARGET_JAVA
}

DomElement *WWebWidget::renderRemove()
{
  DomElement *e = DomElement::getForUpdate(this, DomElement_DIV);
  e->removeFromParent();
  return e;
}

void WWebWidget::removeChild(WWidget *w)
{
  assert(children_ != 0);

  int i = Utils::indexOf(*children_, w);

  assert (i != -1);
    
  if (!flags_.test(BIT_IGNORE_CHILD_REMOVES)
      && !flags_.test(BIT_BEING_DELETED)) {
    DomElement *e = w->webWidget()->renderRemove();

    if (e) {
      if (!transientImpl_)
	transientImpl_ = new TransientImpl();

      transientImpl_->childRemoveChanges_.push_back(e);

      repaint(RepaintInnerHtml);
    }
  }

  /*
    -- does not work properly: should in reality propagate the render
    remove to all grand children; but perhaps we don't need this

    std::vector<DomElement *> *nestedRemoveChanges
      = w->webWidget()->childRemoveChanges_;

    if (nestedRemoveChanges_) {
      for (unsigned k = 0; k < w->nestedRemoveChanges_->size(); ++k) {
	DomElement *f = (*nestedRemoveChanges_)[k];

	if (!f->discardWithParent()) {
	  if (!childRemoveChanges_)
	    childRemoveChanges_ = new std::vector<DomElement *>;
	  childRemoveChanges_->push_back(f);

	  nestedRemoveChanges_->erase(nestedRemoveChanges_->begin() + k);
	  --k;
	}
      }
    }
    */

  w->WObject::setParent((WObject *)0);
    
  /*
   * When the child is about to be deleted, all of its descendants
   * properly removes itself from the renderer "dirty" list. If not,
   * we here force this propagation.
   */
  if (!w->webWidget()->flags_.test(BIT_BEING_DELETED))
    w->webWidget()->quickPropagateRenderOk();

  children_->erase(children_->begin() + i);

  WApplication::instance()
    ->session()->renderer().updateFormObjects(w->webWidget(), true);
}

void WWebWidget::setPositionScheme(PositionScheme scheme)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->positionScheme_ = scheme;

  if ((scheme == Absolute) || (scheme == Fixed))
    flags_.reset(BIT_INLINE);

  flags_.set(BIT_GEOMETRY_CHANGED);
  repaint(RepaintPropertyAttribute);
}

PositionScheme WWebWidget::positionScheme() const
{
  return layoutImpl_ ? layoutImpl_->positionScheme_ : Static;
}

void WWebWidget::resize(const WLength& width, const WLength& height)
{
  if (!width_ && !width.isAuto())
    width_ = new WLength();

  if (width_)
    *width_ = width;

  if (!height_ && !height.isAuto())
    height_ = new WLength();

  if (height_)
    *height_ = height;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

WLength WWebWidget::width() const
{
  return width_ ? *width_ : WLength::Auto;
}

WLength WWebWidget::height() const
{
  return height_ ? *height_ : WLength::Auto;
}

void WWebWidget::setMinimumSize(const WLength& width, const WLength& height)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->minimumWidth_ = width;
  layoutImpl_->minimumHeight_ = height;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

WLength WWebWidget::minimumWidth() const
{
  return layoutImpl_ ? layoutImpl_->minimumWidth_ : WLength::Auto;
}

WLength WWebWidget::minimumHeight() const
{
  return layoutImpl_ ? layoutImpl_->minimumHeight_ : WLength::Auto;
}

void WWebWidget::setMaximumSize(const WLength& width, const WLength& height)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->maximumWidth_ = width;
  layoutImpl_->maximumHeight_ = height;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

WLength WWebWidget::maximumWidth() const
{
  return layoutImpl_ ? layoutImpl_->maximumWidth_ : WLength::Auto;
}

WLength WWebWidget::maximumHeight() const
{
  return layoutImpl_ ? layoutImpl_->maximumHeight_ : WLength::Auto;
}

void WWebWidget::setLineHeight(const WLength& height)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->lineHeight_ = height;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

WLength WWebWidget::lineHeight() const
{
  return layoutImpl_ ? layoutImpl_->lineHeight_ : WLength::Auto;
}

#ifndef WT_TARGET_JAVA
void WWebWidget::setFloatSide(Side s)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->floatSide_ = s;

  flags_.set(BIT_FLOAT_SIDE_CHANGED);

  repaint(RepaintPropertyAttribute);
}

Side WWebWidget::floatSide() const
{
  if (layoutImpl_)
    return layoutImpl_->floatSide_;
  else
    return static_cast<Side>(0);
}

void WWebWidget::setClearSides(WFlags<Side> sides)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->clearSides_ = sides;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

WFlags<Side> WWebWidget::clearSides() const
{
  if (layoutImpl_)
    return layoutImpl_->clearSides_;
  else
    return WFlags<Side>(None);
}
#endif

void WWebWidget::setVerticalAlignment(AlignmentFlag va,
				      const WLength& length)
{
#ifndef WT_TARGET_JAVA // cnor fix this
  if (va & AlignHorizontalMask) {
    wApp->log("warning") << "WWebWidget::setVerticalAlignment(): alignment "
      "(" << va << ") is horizontal, expected vertical";
    va = AlignmentFlag(va & AlignVerticalMask);
  }
#endif // WT_TARGET_JAVA
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->verticalAlignment_ = va;
  layoutImpl_->verticalAlignmentLength_ = length;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

AlignmentFlag WWebWidget::verticalAlignment() const
{
  return layoutImpl_ ? layoutImpl_->verticalAlignment_ : AlignBaseline;
}

WLength WWebWidget::verticalAlignmentLength() const
{
  return layoutImpl_ ? layoutImpl_->verticalAlignmentLength_ : WLength::Auto;
}

void WWebWidget::setOffsets(const WLength& length, WFlags<Side> sides)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();  

  if (sides & Top)
    layoutImpl_->offsets_[0] = length;
  if (sides & Right)
    layoutImpl_->offsets_[1] = length;
  if (sides & Bottom)
    layoutImpl_->offsets_[2] = length;
  if (sides & Left)
    layoutImpl_->offsets_[3] = length;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

WLength WWebWidget::offset(Side s) const
{
  if (!layoutImpl_)
    return WLength::Auto;

  switch (s) {
  case Top:
    return layoutImpl_->offsets_[0];
  case Right:
    return layoutImpl_->offsets_[1];
  case Bottom:
    return layoutImpl_->offsets_[2];
  case Left:
    return layoutImpl_->offsets_[3];
  default:
    throw WtException("WWebWidget::offset(Side) with invalid side.");
  }
}

int WWebWidget::zIndex() const
{
  if (layoutImpl_)
    return layoutImpl_->popup_ ? 300 : 0;
  else
    return 0;
}

void WWebWidget::setInline(bool inl)
{
  flags_.set(BIT_INLINE, inl);

  resetLearnedSlot(&WWidget::show);

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

bool WWebWidget::isInline() const
{
  return flags_.test(BIT_INLINE);
}

void WWebWidget::setPopup(bool popup)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();

  layoutImpl_->popup_ = popup;

  flags_.set(BIT_GEOMETRY_CHANGED);

  repaint(RepaintPropertyAttribute);
}

bool WWebWidget::isPopup() const
{
  return layoutImpl_ ? layoutImpl_->popup_ : false;
}


void WWebWidget::setMargin(const WLength& length, WFlags<Side> sides)
{
  if (!layoutImpl_)
    layoutImpl_ = new LayoutImpl();  

  if (sides & Top)
    layoutImpl_->margin_[0] = length;
  if (sides & Right)
    layoutImpl_->margin_[1] = length;
  if (sides & Bottom)
    layoutImpl_->margin_[2] = length;
  if (sides & Left)
    layoutImpl_->margin_[3] = length;

  layoutImpl_->marginsChanged_ = true;

  repaint(RepaintPropertyAttribute);
}

WLength WWebWidget::margin(Side side) const
{
  if (!layoutImpl_)
    return WLength(0);

  switch (side) {
  case Top:
    return layoutImpl_->margin_[0];
  case Right:
    return layoutImpl_->margin_[1];
  case Bottom:
    return layoutImpl_->margin_[2];
  case Left:
    return layoutImpl_->margin_[3];
  default:
    throw WtException("WWebWidget::margin(Side) with invalid side");
  }
}

void WWebWidget::setStyleClass(const WT_USTRING& styleClass)
{
  if (!lookImpl_)
    lookImpl_ = new LookImpl();

  lookImpl_->styleClass_ = styleClass;
  lookImpl_->styleClassChanged_ = true;

  repaint(RepaintPropertyAttribute);
}

void WWebWidget::setStyleClass(const char *styleClass)
{
  setStyleClass(WString::fromUTF8(styleClass));
}

WT_USTRING WWebWidget::styleClass() const
{
  return lookImpl_ ? lookImpl_->styleClass_ : WT_USTRING();
}

void WWebWidget::setAttributeValue(const std::string& name,
				   const WT_USTRING& value)
{
  if (!otherImpl_)
    otherImpl_ = new OtherImpl();

  if (!otherImpl_->attributes_)
    otherImpl_->attributes_ = new std::map<std::string, WT_USTRING>;
  (*otherImpl_->attributes_)[name] = value;

  if (!otherImpl_->attributesSet_)
    otherImpl_->attributesSet_ = new std::vector<std::string>;

  otherImpl_->attributesSet_->push_back(name);

  repaint(RepaintPropertyAttribute);
}

void WWebWidget::setToolTip(const WString& message)
{
  if (canOptimizeUpdates() && (toolTip() == message))
    return;

  if (!lookImpl_)
    lookImpl_ = new LookImpl();

  if (!lookImpl_->toolTip_)
    lookImpl_->toolTip_ = new WString();

  *lookImpl_->toolTip_ = message;

  lookImpl_->toolTipChanged_ = true;

  repaint(RepaintPropertyAttribute);
}

WString WWebWidget::toolTip() const
{
  return lookImpl_ ? (lookImpl_->toolTip_ ? *lookImpl_->toolTip_ : WString())
    : WString();
}

void WWebWidget::setHidden(bool hidden)
{
  if (canOptimizeUpdates() && (hidden == isHidden()))
    return;

  if (hidden)
    flags_.set(BIT_HIDDEN);
  else
    flags_.reset(BIT_HIDDEN);

  flags_.set(BIT_HIDDEN_CHANGED);

  WApplication::instance()
    ->session()->renderer().updateFormObjects(this, true);

  repaint(RepaintPropertyAttribute);
}

bool WWebWidget::isHidden() const
{
  return flags_.test(BIT_HIDDEN);
}

void WWebWidget::addChild(WWidget *child)
{
  if (child->parent() != 0) {
    child->setParent(0);
    wApp->log("warn") << "WWebWidget::addChild(): reparenting child";
  }

  if (!children_)
    children_ = new std::vector<WWidget *>;

  children_->push_back(child);

  child->WObject::setParent((WObject *)this);

  if (flags_.test(BIT_LOADED))
    doLoad(child);

  WApplication::instance()
    ->session()->renderer().updateFormObjects(this, false);
}

const std::vector<WWidget *>& WWebWidget::children() const
{
  return children_ ? *children_ : emptyWidgetList_;
}

void WWebWidget::setHideWithOffsets(bool how)
{
  if (how) {
    if (!flags_.test(BIT_HIDE_WITH_OFFSETS)) {
      flags_.set(BIT_HIDE_WITH_OFFSETS);

      resetLearnedSlot(&WWidget::show);
      resetLearnedSlot(&WWidget::hide);

      if (parent())
	parent()->setHideWithOffsets(true);
    }
  }
}

void WWebWidget::updateDom(DomElement& element, bool all)
{
  /*
   * determine display
   */
  if (flags_.test(BIT_GEOMETRY_CHANGED)
      || (!flags_.test(BIT_HIDE_WITH_OFFSETS)
	  && flags_.test(BIT_HIDDEN_CHANGED))
      || all) {
    if (flags_.test(BIT_HIDE_WITH_OFFSETS) || !flags_.test(BIT_HIDDEN)) {
      if (element.isDefaultInline() != flags_.test(BIT_INLINE)) {
	if (flags_.test(BIT_INLINE)) {
	  if (element.type() == DomElement_TABLE)
	    element.setProperty(PropertyStyleDisplay, "inline-table");
	  if (element.type() == DomElement_LI)
	    element.setProperty(PropertyStyleDisplay, "inline");
	  else if (element.type() != DomElement_TD)
	    element.setProperty(PropertyStyleDisplay, "inline-block");
	} else {
	  element.setProperty(PropertyStyleDisplay, "block");
	}
      } else if (!all && flags_.test(BIT_HIDDEN_CHANGED))
	if (element.isDefaultInline() == flags_.test(BIT_INLINE))
	  element.setProperty(PropertyStyleDisplay, "");
	else
	  element.setProperty(PropertyStyleDisplay,
			      flags_.test(BIT_INLINE) ? "inline" : "block");
    } else
      element.setProperty(PropertyStyleDisplay, "none");

    if (!flags_.test(BIT_HIDE_WITH_OFFSETS))
      flags_.reset(BIT_HIDDEN_CHANGED);
  }

  if (flags_.test(BIT_GEOMETRY_CHANGED) || all) {
    if (layoutImpl_) {
      /*
       * set position
       */
      switch (layoutImpl_->positionScheme_) {
      case Static:
	break;
      case Relative:
	element.setProperty(PropertyStylePosition, "relative"); break;
      case Absolute:
	element.setProperty(PropertyStylePosition, "absolute"); break;
      case Fixed:
	element.setProperty(PropertyStylePosition, "fixed"); break;
      }

      /*
       * set z-index
       */
      if (layoutImpl_->popup_) {
	element.setProperty(PropertyStyleZIndex,
			    boost::lexical_cast<std::string>(zIndex()));
	WApplication *app = WApplication::instance();
	if (all && app->environment().agentIE6()) {
	  DomElement *i = DomElement::createNew(DomElement_IFRAME);
	  i->setId("sh" + id());
	  i->setAttribute("class", "Wt-shim");
	  i->setAttribute("src", "javascript:false;");
	  i->setAttribute("title", "Popup Shim");
	  i->setAttribute("tabindex", "-1");
	  i->setAttribute("frameborder", "0");

	  app->addAutoJavaScript
	    ("{var w = " + jsRef() + ";"
	     "if (w && !" WT_CLASS ".isHidden(w)) {"
	     "var i = " WT_CLASS ".getElement('" + i->id() + "');"
	     "i.style.width=w.clientWidth + 'px';"
	     "i.style.height=w.clientHeight + 'px';"
	     "}}");

	  element.addChild(i);
	}
      }

#ifndef WT_TARGET_JAVA
      /*
       * set clear: FIXME: multiple values
       */
      if (layoutImpl_->clearSides_ == Left) {
	element.setProperty(PropertyStyleClear, "left");
      } else if (layoutImpl_->clearSides_ == Right) {
	element.setProperty(PropertyStyleClear, "right");
      } else if (layoutImpl_->clearSides_ == Verticals) {
	element.setProperty(PropertyStyleClear, "both");
      }
#endif // WT_TARGET_JAVA

      if (!layoutImpl_->minimumWidth_.isAuto()
	  && (layoutImpl_->minimumWidth_.value() != 0))
	element.setProperty(PropertyStyleMinWidth,
			    layoutImpl_->minimumWidth_.cssText());
      if (!layoutImpl_->minimumHeight_.isAuto()
	  && (layoutImpl_->minimumHeight_.value() != 0))
	element.setProperty(PropertyStyleMinHeight,
			    layoutImpl_->minimumHeight_.cssText());
      if (!layoutImpl_->maximumWidth_.isAuto()) // == none
	element.setProperty(PropertyStyleMaxWidth,
			    layoutImpl_->maximumWidth_.cssText());
      if (!layoutImpl_->maximumHeight_.isAuto()) // == none
	element.setProperty(PropertyStyleMaxHeight,
			    layoutImpl_->maximumHeight_.cssText());

      /*
       * set offsets
       */
      if (layoutImpl_->positionScheme_ != Static) {
	static const Property properties[] = { PropertyStyleTop,
					       PropertyStyleRight,
					       PropertyStyleBottom,
					       PropertyStyleLeft };

	for (unsigned i = 0; i < 4; ++i) {
	  if (!layoutImpl_->offsets_[i].isAuto())
	    element.setProperty(properties[i],
				layoutImpl_->offsets_[i].cssText());
	}
      }

      /*
       * set vertical alignment
       */
      switch (layoutImpl_->verticalAlignment_) {
      case AlignBaseline:
	break;
      case AlignSub:
	element.setProperty(PropertyStyleVerticalAlign, "sub"); break;
      case AlignSuper:
	element.setProperty(PropertyStyleVerticalAlign, "super"); break;
      case AlignTop:
	element.setProperty(PropertyStyleVerticalAlign, "top"); break;
      case AlignTextTop:
	element.setProperty(PropertyStyleVerticalAlign, "text-top"); break;
      case AlignMiddle:
	element.setProperty(PropertyStyleVerticalAlign, "middle"); break;
      case AlignBottom:
	element.setProperty(PropertyStyleVerticalAlign, "bottom"); break;
      case AlignTextBottom:
	element.setProperty(PropertyStyleVerticalAlign, "text-bottom"); break;
      case AlignLength:
	element.setProperty(PropertyStyleVerticalAlign,
			    layoutImpl_->verticalAlignmentLength_.cssText());
	break;
      default:
	break;
      }

      if (!layoutImpl_->lineHeight_.isAuto()) // == none
	element.setProperty(PropertyStyleLineHeight,
			    layoutImpl_->lineHeight_.cssText());

    }

    /*
     * set width & height
     */
    if (width_ && !width_->isAuto())
      element.setProperty(PropertyStyleWidth, width_->cssText());
    if (height_ && !height_->isAuto())
      element.setProperty(PropertyStyleHeight, height_->cssText());

    flags_.reset(BIT_GEOMETRY_CHANGED);
  }


#ifndef WT_TARGET_JAVA
  if (flags_.test(BIT_FLOAT_SIDE_CHANGED) || all) {
    if (layoutImpl_) {
      /*
       * set float
       */
      switch (layoutImpl_->floatSide_) {
      case 0:
	if (flags_.test(BIT_FLOAT_SIDE_CHANGED))
	  element.setProperty(PropertyStyleFloat, "none");
	break;
      case Left:
	element.setProperty(PropertyStyleFloat, "left");
	break;
      case Right:
	element.setProperty(PropertyStyleFloat, "right");
	break;
      default:
	/* illegal values */
	;
      }
    }

    flags_.reset(BIT_FLOAT_SIDE_CHANGED);
  }
#endif // WT_TARGET_JAVA

  if (layoutImpl_) {
    if (layoutImpl_->marginsChanged_ || all) {
      if (layoutImpl_->marginsChanged_
	  || (layoutImpl_->margin_[0].value() != 0))
	element.setProperty(PropertyStyleMarginTop,
			    layoutImpl_->margin_[0].cssText());
      if (layoutImpl_->marginsChanged_
	  || (layoutImpl_->margin_[1].value() != 0))
	element.setProperty(PropertyStyleMarginRight,
			    layoutImpl_->margin_[1].cssText());
      if (layoutImpl_->marginsChanged_
	  || (layoutImpl_->margin_[2].value() != 0))
	element.setProperty(PropertyStyleMarginBottom,
			    layoutImpl_->margin_[2].cssText());
      if (layoutImpl_->marginsChanged_
	  || (layoutImpl_->margin_[3].value() != 0))
	element.setProperty(PropertyStyleMarginLeft,
			    layoutImpl_->margin_[3].cssText());

      layoutImpl_->marginsChanged_ = false;
    }
  }

  if (lookImpl_) {
    if (lookImpl_->toolTip_
	&& (lookImpl_->toolTipChanged_ || all)) {
      if ((lookImpl_->toolTip_->value().length() > 0)
	  || lookImpl_->toolTipChanged_)
	element.setAttribute("title", lookImpl_->toolTip_->toUTF8());

      lookImpl_->toolTipChanged_ = false;
    }

    if (lookImpl_->decorationStyle_)
      lookImpl_->decorationStyle_->updateDomElement(element, all);

    if (((!all) && lookImpl_->styleClassChanged_)
	|| (all && !lookImpl_->styleClass_.empty()))
      element.setAttribute("class", lookImpl_->styleClass_.toUTF8());

    lookImpl_->styleClassChanged_ = false;
  }

  if (otherImpl_ && otherImpl_->attributes_) {
    if (all) {
      for (std::map<std::string, WT_USTRING>::const_iterator i
	     = otherImpl_->attributes_->begin();
	   i != otherImpl_->attributes_->end(); ++i)
	element.setAttribute(i->first, i->second.toUTF8());
    } else if (otherImpl_->attributesSet_) {
      for (unsigned i = 0; i < otherImpl_->attributesSet_->size(); ++i) {
	std::string attr = (*otherImpl_->attributesSet_)[i];
	element.setAttribute(attr, (*otherImpl_->attributes_)[attr].toUTF8());
      }
    }

    delete otherImpl_->attributesSet_;
    otherImpl_->attributesSet_ = 0;
  }

  if (flags_.test(BIT_HIDE_WITH_OFFSETS)) {
    if (flags_.test(BIT_HIDDEN_CHANGED)
	|| (all && flags_.test(BIT_HIDDEN))) {

      if (flags_.test(BIT_HIDDEN)) {
	element.setProperty(PropertyStylePosition, "absolute");
	element.setProperty(PropertyStyleLeft, "-10000px");
	element.setProperty(PropertyStyleTop, "-10000px");
	element.setProperty(PropertyStyleVisibility, "hidden");
      } else {
	if (layoutImpl_) {
	  switch (layoutImpl_->positionScheme_) {
	  case Static:
	    element.setProperty(PropertyStylePosition, "static"); break;
	  case Relative:
	    element.setProperty(PropertyStylePosition, "relative"); break;
	  case Absolute:
	    element.setProperty(PropertyStylePosition, "absolute"); break;
	  case Fixed:
	    element.setProperty(PropertyStylePosition, "fixed"); break;
	  }
	  element.setProperty(PropertyStyleTop,
			      layoutImpl_->offsets_[0].cssText());
	  element.setProperty(PropertyStyleLeft,
			      layoutImpl_->offsets_[3].cssText());
	} else {
	  element.setProperty(PropertyStylePosition, "static");
	}
	element.setProperty(PropertyStyleVisibility, "visible");
	element.setProperty(PropertyStyleTop, "0px");
	element.setProperty(PropertyStyleLeft, "0px");
	element.setProperty(PropertyStyleDisplay, ""); // XXX
      }

      flags_.reset(BIT_HIDDEN_CHANGED);
    }
  }

  renderOk();

  delete transientImpl_;
  transientImpl_ = 0;
}

bool WWebWidget::isStubbed() const
{
  if (flags_.test(BIT_STUBBED))
    return true;
  else {
    WWidget *p = parent();
    return p ? p->isStubbed() : false;
  }
}

bool WWebWidget::isVisible() const
{
  if (flags_.test(BIT_STUBBED) || flags_.test(BIT_HIDDEN))
    return false;
  else
    if (parent())
      return parent()->isVisible();
    else
      return true;
}

void WWebWidget::getSFormObjects(std::vector<WObject *>& result)
{
  if (!flags_.test(BIT_STUBBED) && !flags_.test(BIT_HIDDEN))
    getFormObjects(result);
}

void WWebWidget::getFormObjects(std::vector<WObject *>& formObjects)
{
  if (flags_.test(BIT_FORM_OBJECT))
    formObjects.push_back(this);

  if (children_)
    for (unsigned i = 0; i < children_->size(); ++i)
      (*children_)[i]->webWidget()->getSFormObjects(formObjects);
}

void WWebWidget::getDomChanges(std::vector<DomElement *>& result,
			       WApplication *app)
{
  DomElement *e = DomElement::getForUpdate(this, domElementType());
  updateDom(*e, false);
  result.push_back(e);
}

void WWebWidget::getSDomChanges(std::vector<DomElement *>& result,
				WApplication *app)
{
  bool isIEMobile = app->environment().agentIEMobile();

  if (flags_.test(BIT_STUBBED)) {
    /*
     * If we are prelearning, we still want to catch changes to *this*
     * widget, since it could be related to the prelearning. Note that
     * this assumes that getDomChanges() does not attempt to do things
     * like recreating the widget...
     *
     * ... which is what happens in WContainerWidget, we make the exception
     * there...
     */
    if (app->session()->renderer().preLearning()) {
      getDomChanges(result, app);
      repaint();
    } else {
      flags_.reset(BIT_STUBBED);

      if (!isIEMobile) {
	DomElement *stub = DomElement::getForUpdate(this, DomElement_SPAN);
	render();
	DomElement *realElement = createDomElement(app);
	stub->replaceWith(realElement, !flags_.test(BIT_HIDE_WITH_OFFSETS));
	result.push_back(stub);
      } else
	propagateRenderOk();
    }
  } else {
    render();

    if (isIEMobile) {
      if (flags_.test(BIT_REPAINT_PROPERTY_ATTRIBUTE)) {
	WWidget *p = this;
	WWebWidget *w = this;
	do {
	  p = p->parent();
	  if (p)
	    w = p->webWidget();
	} while (p && w == this);

	w->getSDomChanges(result, app);
      } else if (flags_.test(BIT_REPAINT_INNER_HTML)
		 || !flags_.test(BIT_REPAINT_PROPERTY_IEMOBILE)) {
	// the last condition results from repainting the parent, in which
	// case no change flags are set
	DomElement *e = createDomElement(app);
	e->updateInnerHtmlOnly();
	result.push_back(e);
      } else { // BIT_REPAINT_PROPERTY_IEMOBILE only
	getDomChanges(result, app);
      }

      return;
    }

    if (transientImpl_) {
      Utils::insert(result, transientImpl_->childRemoveChanges_);
      transientImpl_->childRemoveChanges_.clear();
    }

    getDomChanges(result, app);
  }
}

void WWebWidget::doneRerender()
{
  if (children_)
    for (unsigned i = 0; i < children_->size(); ++i)
      (*children_)[i]->webWidget()->doneRerender();
}

void WWebWidget::propagateRenderOk(bool deep)
{
  if (flags_.test(BIT_STUBBED))
    return;

  if (needRerender()) {
    DomElement *v = DomElement::createNew(DomElement_SPAN);
    updateDom(*v, false);
    delete v;
  }

  if (deep && children_)
    for (unsigned i = 0; i < children_->size(); ++i)
      (*children_)[i]->webWidget()->propagateRenderOk();

  delete transientImpl_;
  transientImpl_ = 0;
}

void WWebWidget::quickPropagateRenderOk()
{
  renderOk();

  if (children_)
    for (unsigned i = 0; i < children_->size(); ++i)
      (*children_)[i]->webWidget()->quickPropagateRenderOk();
}

void WWebWidget::setLoadLaterWhenInvisible(bool how)
{
  flags_.set(BIT_DONOT_STUB, !how);
}

void WWebWidget::setId(DomElement *element, WApplication *app)
{
  if (!app->environment().agentIsSpiderBot()
      || (otherImpl_ && otherImpl_->id_))
    element->setId(this, flags_.test(BIT_FORM_OBJECT));
}

DomElement *WWebWidget::createDomElement(WApplication *app)
{
  DomElement *result = DomElement::createNew(domElementType());
  setId(result, app);

  updateDom(*result, true);

  return result;
}

DomElement *WWebWidget::createSDomElement(WApplication *app)
{
  if (!flags_.test(BIT_DONOT_STUB)
      && flags_.test(BIT_HIDDEN)
      && WApplication::instance()->session()->renderer().visibleOnly()) {
    quickPropagateRenderOk();

    flags_.set(BIT_STUBBED);

    DomElement *stub = DomElement::createNew(DomElement_SPAN);
    if (!flags_.test(BIT_HIDE_WITH_OFFSETS)) {
      stub->setProperty(Wt::PropertyStyleDisplay, "none");
    } else {
      stub->setProperty(PropertyStylePosition, "absolute");
      stub->setProperty(PropertyStyleLeft, "-10000px");
      stub->setProperty(PropertyStyleTop, "-10000px");
      stub->setProperty(PropertyStyleVisibility, "hidden");
    }
    if (wApp->environment().javaScript())
      stub->setProperty(Wt::PropertyInnerHTML, "...");

    if (!app->environment().agentIsSpiderBot()
	|| (otherImpl_ && otherImpl_->id_))
      stub->setId(this);

    repaint();

    return stub;
  } else {
    flags_.reset(BIT_STUBBED);

    render();

    return createDomElement(app);
  }
}

void WWebWidget::refresh()
{
  if (lookImpl_ && lookImpl_->toolTip_)
    if (lookImpl_->toolTip_->refresh()) {
      lookImpl_->toolTipChanged_ = true;
      repaint(RepaintPropertyAttribute);
    }

  if (children_)
    for (unsigned i = 0; i < children_->size(); ++i)
      (*children_)[i]->refresh();
}

void WWebWidget::setIgnoreChildRemoves(bool how)
{
  if (how)
    flags_.set(BIT_IGNORE_CHILD_REMOVES);
  else
    flags_.reset(BIT_IGNORE_CHILD_REMOVES);
}

WString WWebWidget::escapeText(const WString& text, bool newlinestoo)
{
  std::string result = text.toUTF8();
  result = escapeText(result, newlinestoo);
  return WString::fromUTF8(result);
}

std::string& WWebWidget::escapeText(std::string& text, bool newlinestoo)
{
  text = Wt::Utils::escapeText(text, newlinestoo);
  return text;
}

std::string WWebWidget::jsStringLiteral(const std::string& value,
					char delimiter)
{
  std::stringstream result;
  DomElement::jsStringLiteral(result, value, delimiter);
  return result.str();
}

std::string WWebWidget::jsStringLiteral(const WString& value,
					char delimiter)
{
  return value.jsStringLiteral(delimiter);
}

void WWebWidget::load()
{
  flags_.set(BIT_LOADED);

  if (children_)
    for (unsigned i = 0; i < children_->size(); ++i)
      doLoad((*children_)[i]);

  if (flags_.test(BIT_HIDE_WITH_OFFSETS))
    parent()->setHideWithOffsets(true);
}

void WWebWidget::doLoad(WWidget *w)
{
  w->load();
  if (!w->loaded())
    std::cerr << "Improper load() implementation: base implementation not "
      "called?" << std::endl;
}

bool WWebWidget::loaded() const
{
  return flags_.test(BIT_LOADED);
}

WWebWidget::DropMimeType::DropMimeType(const WT_USTRING& aHoverStyleClass)
  : hoverStyleClass(aHoverStyleClass)
{ }

WWebWidget::DropMimeType::DropMimeType()
{ }

bool WWebWidget::setAcceptDropsImpl(const std::string& mimeType, bool accept,
				    const WT_USTRING& hoverStyleClass)
{
  bool result  = false; // whether the signal needs to be connected.
  bool changed = false;

  if (!otherImpl_)
    otherImpl_ = new OtherImpl();
  if (!otherImpl_->acceptedDropMimeTypes_)
    otherImpl_->acceptedDropMimeTypes_ = new OtherImpl::MimeTypesMap;

  OtherImpl::MimeTypesMap::iterator i
    = otherImpl_->acceptedDropMimeTypes_->find(mimeType);

  if (i == otherImpl_->acceptedDropMimeTypes_->end()) {
    if (accept) {
      result = otherImpl_->acceptedDropMimeTypes_->empty();
      (*otherImpl_->acceptedDropMimeTypes_)[mimeType]
	= DropMimeType(hoverStyleClass);
      changed = true;
    }
  } else {
    if (!accept) {
      otherImpl_->acceptedDropMimeTypes_->erase(i);
      changed = true;
    }
  }

  if (changed) {
    std::string mimeTypes = "";

    for (OtherImpl::MimeTypesMap::const_iterator j
	   = otherImpl_->acceptedDropMimeTypes_->begin();
	 j != otherImpl_->acceptedDropMimeTypes_->end(); ++j) {
      mimeTypes
	+= "{" + j->first + ":" + j->second.hoverStyleClass.toUTF8() + "}";
    }

    setAttributeValue("amts", mimeTypes);
  }

  if (result && !otherImpl_->dropSignal_)
    otherImpl_->dropSignal_
      = new JSignal<std::string,std::string, WMouseEvent>(this, "_drop");

  return result;
}

EventSignal<WKeyEvent> *WWebWidget::keyEventSignal(const char *name,
						   bool create)
{
  EventSignalBase *b = getEventSignal(name);
  if (b)
    return static_cast<EventSignal<WKeyEvent> *>(b);
  else if (!create)
    return 0;
  else {
#ifndef WT_TARGET_JAVA
    EventSignal<WKeyEvent> *result = new EventSignal<WKeyEvent>(name, this);
#else
    EventSignal<WKeyEvent> *result
      = new EventSignal<WKeyEvent>(name, this, WKeyEvent::templateEvent);
#endif // WT_TARGET_JAVA

    addEventSignal(*result);
    return result;
  }
}

EventSignal<void> *WWebWidget::voidEventSignal(const char *name, bool create)
{
  EventSignalBase *b = getEventSignal(name);
  if (b)
    return static_cast<EventSignal<void> *>(b);
  else if (!create)
    return 0;
  else {
    EventSignal<void> *result = new EventSignal<void>(name, this);
    addEventSignal(*result);
    return result;
  }
}

EventSignal<WMouseEvent> *WWebWidget::mouseEventSignal(const char *name,
						       bool create)
{
  EventSignalBase *b = getEventSignal(name);
  if (b)
    return static_cast<EventSignal<WMouseEvent> *>(b);
  else if (!create)
    return 0;
  else {
#ifndef WT_TARGET_JAVA
    EventSignal<WMouseEvent> *result = new EventSignal<WMouseEvent>(name, this);
#else
    EventSignal<WMouseEvent> *result
      = new EventSignal<WMouseEvent>(name, this, WMouseEvent::templateEvent);
#endif // WT_TARGET_JAVA
    addEventSignal(*result);
    return result;
  }
}

EventSignal<WScrollEvent> *WWebWidget::scrollEventSignal(const char *name,
							 bool create)
{
  EventSignalBase *b = getEventSignal(name);
  if (b)
    return static_cast<EventSignal<WScrollEvent> *>(b);
  else if (!create)
    return 0;
  else {
#ifndef WT_TARGET_JAVA
    EventSignal<WScrollEvent> *result
      = new EventSignal<WScrollEvent>(name, this);
#else
    EventSignal<WScrollEvent> *result
      = new EventSignal<WScrollEvent>(name, this, WScrollEvent::templateEvent);
#endif // WT_TARGET_JAVA
    addEventSignal(*result);
    return result;
  }
}

void WWebWidget::updateSignalConnection(DomElement& element,
					EventSignalBase &signal,
					const char *name,
					bool all)
{
  if (name[0] != 'M' && (all || signal.needUpdate())) {
    if (signal.isConnected())
      element.setEventSignal(name, signal);
    else
      if (!all)
	element.setEvent(name, std::string(), std::string());

    signal.updateOk();
  }
}

bool WWebWidget::canOptimizeUpdates()
{
  return !WApplication::instance()->session()->renderer().preLearning();
}

std::string WWebWidget::fixRelativeUrl(const std::string& url)
{
  return WApplication::instance()->fixRelativeUrl(url);
}

bool WWebWidget::removeScript(WString& text)
{
#ifndef WT_NO_XSS_FILTER
  return XSSFilterRemoveScript(text);
#else
  return true;
#endif
}
