/*
 * Copyright (C) 2008 Emweb bvba, Kessel-Lo, Belgium.
 *
 * See the LICENSE file for terms of use.
 */
#include "Wt/WApplication"
#include "Wt/WWidget"
#include "Wt/WWebWidget"
#include "Wt/WContainerWidget"
#include "Wt/WLayout"
#include "Wt/WJavaScript"

#include "DomElement.h"
#include "EscapeOStream.h"
#include "WebRenderer.h"
#include "WebSession.h"
#include "WtException.h"

namespace Wt {

WWidget::WWidget(WContainerWidget* parent)
  : WObject(0),
    needRerender_(true)
{ }

WWidget::~WWidget()
{
  while (!eventSignals_.empty()) {
    EventSignalBase *s = &eventSignals_.front();
    eventSignals_.pop_front();
#ifndef WT_TARGET_JAVA
    delete s;
#else
    s->~EventSignalBase();
#endif
  }

  if (needRerender_)
    WApplication::instance()->session()->renderer().doneUpdate(this);
}

void WWidget::setParent(WWidget *p)
{
  if (parent())
    parent()->removeChild(this);
  if (p)
    p->addChild(this);
}

void WWidget::render()
{
  renderOk();
}

void WWidget::renderOk()
{
  if (needRerender_) {
    needRerender_ = false;
    WApplication::instance()->session()->renderer().doneUpdate(this);
  }
}

void WWidget::askRerender()
{
  if (!needRerender_) {
    needRerender_ = true;
    WApplication::instance()->session()->renderer().needUpdate(this);
  }
}

void WWidget::setStyleClass(const char *styleClass)
{
  setStyleClass(WString::fromUTF8(styleClass));
}

void WWidget::hide()
{
  wasHidden_ = isHidden();
  setHidden(true);
}

void WWidget::show()
{ 
  wasHidden_ = isHidden();
  setHidden(false);
}

#ifndef WT_TARGET_JAVA
WStatelessSlot *WWidget::getStateless(Method method)
{
  if (method == static_cast<WObject::Method>(&WWidget::hide))
    return implementStateless(&WWidget::hide, &WWidget::undoHideShow);
  else if (method == static_cast<WObject::Method>(&WWidget::show))
    return implementStateless(&WWidget::show, &WWidget::undoHideShow);
  else
    return WObject::getStateless(method);
}
#endif // WT_TARGET_JAVA

void WWidget::undoHideShow()
{
  setHidden(wasHidden_);
}

#ifdef WT_TARGET_JAVA
void WWidget::resize(int widthPixels, int heightPixels)
{
  resize(WLength(widthPixels), WLength(heightPixels));
}

void WWidget::setMargin(int pixels, WFlags<Side> sides)
{
  setMargin(WLength(pixels), sides);
}

void WWidget::setOffsets(int pixels, WFlags<Side> sides)
{
  setOffsets(WLength(pixels), sides);
}
#endif // WT_TARGET_JAVA

std::string WWidget::jsRef() const
{
  return WT_CLASS ".getElement('" + formName() + "')";
}

void WWidget::htmlText(std::ostream& out)
{
  DomElement *element
    = webWidget()->createSDomElement(WApplication::instance());
  DomElement::TimeoutList timeouts;
  EscapeOStream sout(out);
  element->asHTML(sout, timeouts);
  delete element;
}

std::string WWidget::inlineCssStyle()
{
  WWebWidget *ww = webWidget();
  DomElement *e = DomElement::getForUpdate(ww, ww->domElementType());
  ww->updateDom(*e, true);
  std::string result = e->cssStyle();
  delete e;
  return result;
}

WWidget *WWidget::adam()
{
  WWidget *p = parent();
  return p ? p->adam() : this;
}

WString WWidget::tr(const char *key)
{
  return WString::tr(key);
}

void WWidget::acceptDrops(const std::string& mimeType,
			  const WT_USTRING& hoverStyleClass)
{
  WWebWidget *thisWebWidget = webWidget();

  if (thisWebWidget->setAcceptDropsImpl(mimeType, true, hoverStyleClass)) {
    thisWebWidget->otherImpl_
      ->dropSignal_->connect(SLOT(this, WWidget::getDrop));
  }
}

void WWidget::stopAcceptDrops(const std::string& mimeType)
{
  WWebWidget *thisWebWidget = webWidget();

  thisWebWidget->setAcceptDropsImpl(mimeType, false, "");
}

void WWidget::getDrop(const std::string sourceId, const std::string mimeType,
		      WMouseEvent event)
{
  WDropEvent e(WApplication::instance()->decodeObject(sourceId), mimeType,
	       event);

  dropEvent(e);
}

void WWidget::dropEvent(WDropEvent event)
{ }

std::string WWidget::createJavaScript(std::stringstream& js,
				      const std::string& insertJS)
{
  DomElement *de = webWidget()->createSDomElement(WApplication::instance());
  std::string var = de->asJavaScript(js, true);
  if (!insertJS.empty())
    js << insertJS << var << ");";
  de->asJavaScript(js, false);
  delete de;

  return var;
}

void WWidget::setLayout(WLayout *layout)
{ 
  layout->setParent(this);
}

WLayout *WWidget::layout()
{
  return 0;
}

WLayoutItemImpl *WWidget::createLayoutItemImpl(WLayoutItem *layoutItem)
{
  throw WtException("WWidget::setLayout(): widget does not support "
		    "layout managers");
}

void WWidget::addEventSignal(EventSignalBase& s)
{
  eventSignals_.push_back(s);
}

EventSignalBase *WWidget::getEventSignal(const char *name)
{
  for (EventSignalList::iterator i = eventSignals_.begin();
       i != eventSignals_.end(); ++i) {
    EventSignalBase& s = *i;
    if (s.name() == name)
      return &s;
  }

  return 0;
}

}
